{$G+,N+}                                                { 286+, 287+ }
{$R-,S-}

{$IFDEF DPMI}
{$M 65520}
{$ELSE}
{$M 65520,0,655360}
{$ENDIF}

(*************************************************)
(*         !Numeric processor is requred!        *)
(*                                               *)
(*              The Mood Engine v1.1             *)
(*                                               *)
(*        (c) 1999 by Alexei A. Frounze          *)
(*     Homepage: http://www.chat.ru/~alexfru     *)
(*           E-mail: alexfru@chat.ru             *)
(*************************************************)

Uses
  CRT, DOS, Grafix, GfxFont, KeyPres2, Vectors, MapIO, PStack, Misc,
  TMapping;

                               (* CONSTANTS *)

{$I CONSTS.PAS}
Const
  DAng = 2;
  Step = 7;
  PRadius = 16;
  PHeight = 50;
  MethodStr : Array [0..2] of String[53] = (
    'Pas+floating point arithmetic in inner loop.',
    'TAsm+both arithmetics+8-pxl interpolation.',
    'TAsm+both arithmetics+16-pxl interpolation.'
  );

                               (* VARIABLES *)
Const
  PScreen : Pointer = Nil;              { Virtual screen pointer }
  PSBar : Pointer = Nil;                { Status bar bitmap pointer }
Var
  MScale : Integer;                     { Map scale factor }
  Two_d  : Boolean;                     { Map/scene switch }
  Wait4Retrace : Boolean;               { Wait for vertical retrace flag }
  InternalTime,                         { Incremented 18.2*4 times per second }
  T0, T1, Frames,
  TT0, TT1 : LongInt;                   { Time }
  FS : String;                          { FPS string }
  OldQuit : Pointer;                    { Old exit proc }
  Timer : Byte;                         { Internal timer counter }
  OldTimer : Pointer;                   { Old timer handler }
  QuitFlag : Boolean;                   { Wanna quit flag }
  SSBar : Word;                         { Status bar bitmap size }
  LogPal : Boolean;                     { Logarithmic palette flag }

                           (* INIT/DONE FUNCTIONS *)

Procedure SetUpPal;
Begin
  Palette [0,1]:= 0;
  Palette [0,2]:= 0;
  Palette [0,3]:= 0;
  Palette [255, 1] := 15;
  Palette [255, 2] := 63;
  Palette [255, 3] := 15;
  Palette [254, 1] := 15;
  Palette [254, 2] := 15;
  Palette [254, 3] := 63
End;

Function Allocate_Screen : Boolean;
Begin
{  If MaxAvail >= 64000 then Begin
    GetMem (PScreen, 64000);
    SetActivePage (Seg(PScreen^));
    Allocate_Screen := True
  End}
  If MaxAvail >= 64000+16 then Begin    { Pascal's heap manager doesn't }
    GetMem (PScreen, 64000+16);         { always allocate memory at     }
    {$IFDEF DPMI}
    SetActivePage (Seg(PScreen^));
    {$ELSE}
    SetActivePage (Seg(PScreen^)+1);    { the paragraph boundaries  :(( }
    {$ENDIF}
    Allocate_Screen := True
  End
  Else
    Allocate_Screen := False
End;

Procedure DeAllocate_Screen;
Begin
  SetActivePage (VSeg);
{  If PScreen <> Nil then
    FreeMem (PScreen, 64000)}
  If PScreen <> Nil then
    FreeMem (PScreen, 64000+16)
End;

Function Load_StatusBar : Boolean;
Begin
  LoadPCX256 ('STBAR.PCX', PSBar, SSBar, Palette);
  Load_StatusBar := PSBar <> Nil
End;

Procedure DeAllocate_StatusBar;
Begin
  If PSBar <> Nil then
    FreeMem (PSBar, SSBar)
End;

Procedure Handle_Player; Forward;

Procedure NewTimer; Interrupt;
Begin
  Asm DB 66h; PushA End;        { PushAD 'coz we using 32-bit arithmetic }
  If Timer = 0 then
    Asm
      PushF
      Call  DWord Ptr OldTimer
    End
  Else Port[$20] := $20;
  Inc (Timer); If Timer = 4 then Timer := 0;
  Inc (InternalTime);
(* do something *)
  Asm DB 66h; PopA End;         { PopAD 'coz we using 32-bit arithmetic }
End;

Function Find_Sector (Xp, Yp : Double) : Integer; Forward;

Function Init : Boolean;
Var
  R, L : Boolean;
  S : String;
Label
  LEnd;
Begin
  GetIntVec (8, OldTimer);
  R := False;
  ClrScr;

  If ParamCount = 0 then
    L := Load_Map ('MAP.MAP')
  Else
    L := Load_Map (ParamStr(1));
  If not L then Begin
    WriteLn ('Error occured during map or textures files loading.');
    WriteLn ('Check the map file and/or free memory size.');
    Goto LEnd
  End;

  Sector := Find_Sector (Xp, Yp);
  If Sector = -1 then Begin
    WriteLn ('Player position is wrong. No sector has been found for player.');
    WriteLn ('Check the map file.');
    Goto LEnd
  End;

  If not Load_StatusBar then Begin
    WriteLn ('Error occured during status bar loading.');
    Goto LEnd
  End;

  If not Allocate_Screen then Begin
    WriteLn ('Not enough free memory for virtual screen allocation.');
    Goto LEnd
  End;

  If not Allocate_Boundaries then Begin
    WriteLn ('Not enough free memory for boundary table allocation.');
    Goto LEnd
  End;
  Put_Screen_Boundaries;

  InternalTime := 0;
  Timer := 0;
  SetIntVec (8, @NewTimer);
  Port[$43] := $34;                             { Speed up timer      }
  Port[$40] := ($10000 div 4) and $FF;          { counter. Increasing }
  Port[$40] := ($10000 div 4) shr 8;            { frequency up for 4 times. }

  InitGraph;
  ClearScreen (Black);
  SetFont (@Font6x7);
  SetUpPal;
  SetPal (Palette);
  LogPal := False;

  WaitForACSReleased;
  Set_Handler;

  R := True;

  Str (MaxAvail div 1024, S);
  WriteS (0, 0, 0, 255, 'Free: '+S+' KB');
  DisplayPage (VSegA);
  Delay (500);

LEnd:
  Init := R
End;

Procedure Done;
Begin
  Remove_Handler;

{  If IsGraphMode then }CloseGraph;
  DeAllocate_Boundaries;
  DeAllocate_Screen;
  DeAllocate_StatusBar;
  UnLoad_Map
End;

                         (* MISC & MATH FUNCTIONS *)

Function On_The_Left (Var VBeg, VEnd, V : TVertex) : Integer;
Var
  Vdx, Vdy, Pdx, Pdy,
  Result : Double;
Begin
  Vdx := VEnd.x-VBeg.x;
  Vdy := VEnd.y-VBeg.y;
  Pdx := V.x-VBeg.x;
  Pdy := V.y-VBeg.y;
  Result := Vdx*Pdy-Vdy*Pdx;
  If Result < 0 then On_The_Left := -1
  Else If result > 0 then On_The_Left := 1
  Else On_The_Left := 0;
  {  "1" - point to the left of vector vbeg-vend  }
  {  "0" - point on the vector vbeg-vend          }
  { "-1" - point to the right of vector vbeg-vend }
End;

Function On_The_Left2 (Var VBeg, VEnd : TVertex; X, Y : Double) : Integer;
Var
  Vdx, Vdy, Pdx, Pdy,
  Result : Double;
Begin
  Vdx := VEnd.x-VBeg.x;
  Vdy := VEnd.y-VBeg.y;
  Pdx := X-VBeg.x;
  Pdy := Y-VBeg.y;
  Result := Vdx*Pdy-Vdy*Pdx;
  If Result < 0 then On_The_Left2 := -1
  Else If result > 0 then On_The_Left2 := 1
  Else On_The_Left2 := 0;
  {  "1" - point to the left of vector vbeg-vend  }
  {  "0" - point on the vector vbeg-vend          }
  { "-1" - point to the right of vector vbeg-vend }
End;

Function Find_Sector (Xp, Yp : Double) : Integer;
Var
  V : TVertex;
  I, J, F, R : Integer;
Begin
  V.x := Trunc(Xp); V.y := Trunc(Yp);
  For I := 0 to Sectors_Cnt-1 do Begin
    F := 1;
    For J := PSectors^[I].SLineDef to PSectors^[I].ELineDef do Begin
      R := On_The_Left2 (PVertices^[PLineDefs^[J].Vertex1],
                      PVertices^[PLineDefs^[J].Vertex2], Xp, Yp);
      If R <> -1 then F := 0
{      If R = 1 then F := 0}
    End;
    If F <> 0 then Break;
  End;
  If F = 0 then Find_Sector := -1
  Else Find_Sector := I
End;

                           (* DRAWING FUNCTIONS *)

Procedure Draw_Map;
Var
  V1, V2 : Integer;
  x1, y1, x2, y2,
  x1s, z1s, x2s, z2s : Double;
  fang, I : Integer;
Begin
  For I := 0 to LineDefs_Cnt-1 do Begin
    { Let's get the linedef coordinates. }
    V1 := PLineDefs^[I].Vertex1;
    x1 := PVertices^[V1].x;
    y1 := PVertices^[V1].y;

    V2 := PLineDefs^[I].Vertex2;
    x2 := PVertices^[V2].x;
    y2 := PVertices^[V2].y;

    { Let's transform its coordinates to the player relative. }

    x1 := x1-Xp; y1 := y1-Yp;
    x2 := x2-Xp; y2 := y2-Yp;
    fang := Angle-90;
    If fang < 0 then Inc (fang, 360);

    x1s := y1*SinTab[fang] + x1*CosTab[fang];
    z1s := y1*CosTab[fang] - x1*SinTab[fang];
    x2s := y2*SinTab[fang] + x2*CosTab[fang];
    z2s := y2*CosTab[fang] - x2*SinTab[fang];

    { Go ahead! Draw linedefs! }

    If PLineDefs^[I].Typ and 1 = 1 then
      Line2 (Center_X+Trunc(x1s*mscale/16), Center_Y-Trunc(z1s*mscale/16),
             Center_X+Trunc(x2s*mscale/16), Center_Y-Trunc(z2s*mscale/16), 254)
    Else
      Line2 (Center_X+Trunc(x1s*mscale/16), Center_Y-Trunc(z1s*mscale/16),
             Center_X+Trunc(x2s*mscale/16), Center_Y-Trunc(z2s*mscale/16), 255);
    PutPixel2 (Center_X, Center_Y, 255);
  End
End;

Procedure Draw_Sector (SN : Integer);
Type
  TTxt = Array [0..Texture_Size*Texture_Size-1] of Byte;
Var
  V1, V2 : Integer;
  x1, y1, x2, y2,
  x1s, z1s, x2s, z2s, ch, fh : Double;
  fang, I, J : Integer;
  OSN : Integer;
  DF, DC : Integer;
  P : ^TTxt;

Procedure Do_Vertex (NV, NVT : Integer; Height : Double);
Begin
  x1 := PVertices^[NV].x;
  y1 := PVertices^[NV].y;

  x1 := x1-Xp; y1 := y1-Yp;

  x1s := y1*SinTab[fang] + x1*CosTab[fang];
  z1s := y1*CosTab[fang] - x1*SinTab[fang];

  Vertices[NVT].x := x1s;
  Vertices[NVT].y := Height*CosTab[Yang] - z1s*SinTab[Yang];
  Vertices[NVT].z := Height*SinTab[Yang] + z1s*CosTab[Yang];
End;

Begin
  fang := Angle-90;
  If fang < 0 then Inc (fang, 360);
  fh := Zp-PSectors^[SN].FHeight;
  ch := Zp-PSectors^[SN].CHeight;

  { Drawing walls }

  For I := PSectors^[SN].SLineDef to PSectors^[SN].ELineDef do
    If PLineDefs^[I].Typ and 1 = 0 then Begin
      { Let's get the linedef coordinates. }
      V1 := PLineDefs^[I].Vertex1;
      V2 := PLineDefs^[I].Vertex2;

      Do_Vertex (V1, 0, ch);
      Do_Vertex (V2, 1, ch);
      Do_Vertex (V2, 2, fh);
      Do_Vertex (V1, 3, fh);
      LX := PLineDefs^[I].Len;
      LY := fh-ch;
      P := PTextures^[PLineDefs^[I].Texture].P;
      Inc (Word(P), 4);
      T_Map (P^)
    End;

  { Drawing floors }

  J := 0;
  For I := PSectors^[SN].SLineDef to PSectors^[SN].ELineDef do Begin
    If J = 0 then Begin
      Do_Vertex (PLineDefs^[I].Vertex1, J, fh);
      Inc (J)
    End;
    Do_Vertex (PLineDefs^[I].Vertex2, J, fh);
    Inc (J)
  End;
  P := PTextures^[PSectors^[SN].FTexture].P;
  Inc (Word(P), 4);
  T_Map_FC (J, P^, fh, Xp, Yp, Angle, Yang);

  { Drawing ceilings }

  J := 0;
  For I := PSectors^[SN].ELineDef downto PSectors^[SN].SLineDef do Begin
    If J = 0 then Begin
      Do_Vertex (PLineDefs^[I].Vertex2, J, ch);
      Inc (J)
    End;
    Do_Vertex (PLineDefs^[I].Vertex1, J, ch);
    Inc (J)
  End;
  P := PTextures^[PSectors^[SN].CTexture].P;
  Inc (Word(P), 4);
  T_Map_FC (J, P^, ch, Xp, Yp, Angle, Yang);

  { Drawing portals }

  For I := PSectors^[SN].SLineDef to PSectors^[SN].ELineDef do
    If (PLineDefs^[I].Typ and 1 = 1) and
       not Is_On_Stack(PLineDefs^[I].PSector) then Begin

      { Let's find differences between floors' heights and ceilngs' heights }
      OSN := PLineDefs^[I].PSector;
      DF := PSectors^[SN].FHeight - PSectors^[OSN].FHeight;
      DC := PSectors^[SN].CHeight - PSectors^[OSN].CHeight;

      { Let's get the portal linedef coordinates. }
      V1 := PLineDefs^[I].Vertex1;
      V2 := PLineDefs^[I].Vertex2;

      { Drawing wall part over portal }

      If DC > 0 then Begin
        Do_Vertex (V1, 0, ch);
        Do_Vertex (V2, 1, ch);
        Do_Vertex (V2, 2, ch+DC);
        Do_Vertex (V1, 3, ch+DC);
        LX := PLineDefs^[I].Len;
        LY := DC;
        P := PTextures^[PLineDefs^[I].Texture].P;
        Inc (Word(P), 4);
        T_Map (P^)
      End;

      { Drawing wall part under portal }

      If DF < 0 then Begin
        Do_Vertex (V1, 0, fh+DF);
        Do_Vertex (V2, 1, fh+DF);
        Do_Vertex (V2, 2, fh);
        Do_Vertex (V1, 3, fh);
        LX := PLineDefs^[I].Len;
        LY := -DF;
        P := PTextures^[PLineDefs^[I].Texture].P;
        Inc (Word(P), 4);
        T_Map (P^)
      End;

      If DC > 0 then Begin
        Do_Vertex (V1, 0, ch+DC);
        Do_Vertex (V2, 1, ch+DC)
      End
      Else Begin
        Do_Vertex (V1, 0, ch);
        Do_Vertex (V2, 1, ch)
      End;

      If DF < 0 then Begin
        Do_Vertex (V2, 2, fh+DF);
        Do_Vertex (V1, 3, fh+DF)
      End
      Else Begin
        Do_Vertex (V2, 2, fh);
        Do_Vertex (V1, 3, fh)
      End;

      { Projecting portal and drawing neighbouring sector }

      Spush (SN);
      If P_Map then
        Draw_Sector (PLineDefs^[I].PSector);
      SPop
    End
End;

Procedure Draw_View;
Var
  SS : String;
Begin
  ClearScreen (Black);
  If Two_d then
    Draw_Map
  Else Begin
    Draw_Sector (Sector);
    WriteS (0, 0, BLACK, 255, 'TMapping:'+MethodStr[TMethod]);

    If Wait4Retrace then SS := 'On' Else SS := 'Off';
    WriteS (0, 7, BLACK, 255, 'Retrace waiting:'+SS);

    WriteS (0, 14, Black, 255, 'FPS:'+FS);
  End;

  Str (Sector, SS);
  WriteS (0, 21, BLACK, 254, 'Sector:'+SS);

  PutImage (0, 200-34, PSBar^);

  If Wait4Retrace then WaitRetrace;   { Wait for retrace }
  DisplayPage (VSegA)                 { Flip the virtual screen }
End;


Procedure PMove (Stp : Integer; Ang : Integer);
Var
  I, mi, V1, V2 : Integer;
  dx, dy, d, md, x1, y1, wdx, wdy, l, r : Double;
Begin
  If Stp < 0 then Begin
    Stp := -Stp;
    Inc (Ang, 180)
  End;
  If Ang >= 360 then Dec (Ang, 360);
  If Ang < 0 then Inc (Ang, 360);

  dx := Stp*CosTab[Ang];
  dy := Stp*SinTab[Ang];

  mi := -1;
  For I := PSectors^[_Sector].SLineDef to PSectors^[_Sector].ELineDef do Begin
    V1 := PLineDefs^[I].Vertex1; V2 := PLineDefs^[I].Vertex2;
    l := PLineDefs^[I].Len;
    x1 := PVertices^[V1].x;
    y1 := PVertices^[V1].y;
    wdx := PVertices^[V2].x-x1;
    wdy := PVertices^[V2].y-y1;

    d := ((_Xp+dx-x1)*wdy - (_Yp+dy-y1)*wdx) / l;
    r := (wdx*(_Xp+dx-x1) + wdy*(_Yp+dy-y1)) / Sqr(l);

    if (d < pradius) and (r > 0-pradius/l) and (r < 1+pradius/l) then
      If mi = -1 then Begin
        mi := I;
        md := d;
      End
      Else
        If d < md then Begin
          mi := I;
          md := d;
        End
  End;

  If mi <> -1 then Begin { We're near the linedef or we crossed the linedef }
    If (PLineDefs^[mi].Typ and 1 = 1) then Begin        { Check for portal }
      If md < 0 then Begin { We went to the another sector thru the portal }
        _Sector := PLineDefs^[mi].PSector;
        _Zp := PSectors^[_Sector].FHeight + PHeight
      End
    End
    Else Begin                          { We bounced into the wall }
      V1 := PLineDefs^[mi].Vertex1; V2 := PLineDefs^[mi].Vertex2;
      l := PLineDefs^[mi].Len;
      x1 := PVertices^[V1].x;
      y1 := PVertices^[V1].y;
      wdx := PVertices^[V2].x-x1;
      wdy := PVertices^[V2].y-y1;

      r := PRadius - ((_Xp+dx-x1)*wdy - (_Yp+dy-y1)*wdx) / l;
      dx := dx + r*wdy/l;
      dy := dy - r*wdx/l
    End
  End;
  I := Find_Sector (_Xp+dx, _Yp+dy);
  If I <> -1 then Begin                         { Safety!!! }
    _Xp := _Xp + dx;
    _Yp := _Yp + dy;
    _Sector := I
  End
End;

Procedure Quit; Far;
Begin
  SetIntVec (8, OldTimer);
  Port[$43] := $34;                             { Setting usual timer }
  Port[$40] := $00;                             {                     }
  Port[$40] := $00;                             { frequency (18.2 Hz) }
{  If IsGraphMode then }CloseGraph;
  If StackError then
    WriteLn ('Portal Stack overflow/underflow error!'#13#10+
             'Stack can keep only ',Max_Boundaries,' items.');
  ExitProc := OldQuit
End;

Procedure Handle_Player;
Var
  Rot, Mv : Integer;
Begin
    Rot := 0; Mv := 0;
    If KeyMap[sEsc] then QuitFlag := True;         { Esc - exit }

    If (KeyMap[sAlt]) then Begin        { If Alt+Left/Right pressed }
      If (KeyMap[sLeft]) then           { Left - strafe left }
        Inc (Rot);
      If (KeyMap[sRight]) then          { Right - strafe right }
        Dec (Rot);
      If Rot <> 0 then Begin
        PMove (Step div 2, _Angle+Rot*90);       { Actual strafing }
        Rot := 0
      End
    End
    Else Begin
      If KeyMap[sLeft] then Begin       { Left - turn left }
        Inc (_Angle, Dang);
        Rot := 1
      End;
      If KeyMap[sRight] then Begin      { Right - turn right }
        Dec (_Angle, Dang);
        Rot := 1
      End;
      If Rot <> 0 then Begin            { Correct angle }
        If _Angle >= 360 then Dec (_Angle, 360);
        If _Angle < 0 then Inc (_Angle,360);
        Rot := 0
      End
    End;
    If KeyMap[sUp] then                 { Up - move forward }
      Inc (Mv);
    If KeyMap[sDown] then               { Down - move backward }
      Dec (Mv);
    If Mv <> 0 then Begin
      PMove (Mv*step, _Angle);          { Actual moving }
      Mv := 0
    End;
(*
    If KeyMap[sSemiColon] then                  { Increment height }
      _Zp := _Zp + 1;
    If KeyMap[sBackSlash] then                  { Decrement height }
      _Zp := _Zp - 1;
*)
    If KeyMap[sA] then Begin            { Look up }
      Dec (_Yang, DAng);
      If _Yang < 0 then Inc (_Yang, 360);
      If (_Yang > 90) and (_Yang < 270) then _Yang := 270
    End;
    If KeyMap[sZ] then Begin            { Look down }
      Inc (_Yang, DAng);
      If _Yang >= 360 then Dec (_Yang, 360);
      If (_Yang > 90) and (_Yang < 270) then _Yang := 90
    End;
    If KeyMap[sEnd] then                { Look forward }
      _Yang := 0
End;

Procedure Lin2Log;
Var
  I : Word;
  R, G, B, D : Double;
Begin
  If LogPal then Begin
{    D := 63 / Ln(64);}
    D := 63/Sqrt(63);
    For I := 1 to 253 do Begin
      R := Palette[I,1];
      G := Palette[I,2];
      B := Palette[I,3];
{      R := Ln(1+R)*D;
      G := Ln(1+G)*D;
      B := Ln(1+B)*D;}
      R := Sqrt(R)*D;
      G := Sqrt(G)*D;
      B := Sqrt(B)*D;
      SetCRGB (I, Round(R), Round(G), Round (B))
    End;
  End
  Else Begin
    SetUpPal;
    SetPal (Palette)
  End
End;

Label
  LEnd;
                              (* MAIN PROGRAM *)

Begin
  If Test8086 < 2 then Begin
    WriteLn ('80386 or better processor is required.');
    Goto LEnd
  End;
  If Test8087 < 3 then Begin
    WriteLn ('80387 or better numeric processor is required.');
    Goto LEnd
  End;
  If not IsVGA then Begin
    WriteLn ('VGA or SVGA adapeter is required.');
    Goto LEnd
  End;

  If not Init then Begin
    Done;
    Goto LEnd
  End;
  OldQuit := ExitProc;
  ExitProc := @Quit;

  Zp := PSectors^[Sector].FHeight + PHeight;
  MScale := 5;
  YAng := 0;
  Wait4Retrace := True;

  T0 := BIOS_Time;
  T1 := T0;

  Two_d := False;
  QuitFlag := False;
  _Sector := Sector;
  _Xp     := Xp;
  _Yp     := Yp;
  _Zp     := Zp;
  _Angle  := Angle;
  _YAng   := Yang;

  Repeat
    TT0 := InternalTime;
    Draw_View;                          { Show the frame }
    Inc (Frames);

    Handle_Player;

    Sector := _Sector;
    Xp     := _Xp;
    Yp     := _Yp;
    Zp     := _Zp;
    Angle  := _Angle;
    YAng   := _Yang;

    If KeyMap[sTab] then Begin          { Tab - toggle 3d scene and 2d map }
      Two_d := not Two_d;
      While KeyMap[sTab] do
    End;
    If Two_d then Begin
      If (KeyMap[sGrayPlus]) and (mscale < 32) then Inc (MScale);
                                        { "+" (keypad) - make 2d map bigger }
      If (KeyMap[sGrayMinus]) and (mscale > 1) then Dec (MScale);
                                        { "-" (keypad) - make 2d map smaller }
      While (KeyMap[sGrayMinus]) or (KeyMap[sGrayPlus]) do
    End;
    If KeyMap[sV] then Begin            { V - toggle vertical retrace waiting }
      Wait4Retrace := not Wait4Retrace;
      While KeyMap[sV] do
    End;
    If KeyMap[sT] then Begin            { V - toggle vertical retrace waiting }
      Inc (TMethod);
      If TMethod > 2 then TMethod := 0;
      While KeyMap[sT] do
    End;
    If KeyMap[sS] then Begin
      Draw_Sector (Sector);
      SavePCX256 ('MOOD.PCX', 0, 0, Screen_Width, Screen_Height);
      While KeyMap[sS] do
    End;
    If KeyMap[sF11] then Begin
      LogPal := not LogPal;
      Lin2Log;
      While KeyMap[sF11] do
    End;
    T1 := BIOS_Time;
    If T1 - T0 >= 18 then Begin
      Str (Frames:3, FS);
      T0 := T1;
      Frames := 0
    End;

    Repeat                      { Timer synchronization }
      TT1 := InternalTime;
      { Do something }
    Until TT1-TT0 >= 2;         { 2*18.2 = 36.4 times -> 36.4 fps max }

  Until QuitFlag;

  Done;
LEnd:
  WriteLn ('Ŀ');
  WriteLn ('              The Mood Engine v1.1             ');
  WriteLn ('        (c) 1999 by Alexei A. Frounze          ');
  WriteLn ('     Homepage: http://www.chat.ru/~alexfru     ');
  WriteLn ('            E-mail: alexfru@chat.ru            ');
  WriteLn ('')
End.
