{$G+,N+}                                                { 286+, 287+ }
{$R-,S-}
Unit PStack;

Interface

{$I CONSTS.PAS}

Type
  TBoundary = Record
    Ymin, Ymax,
    Xmin, Xmax : Integer;
    Boundary : Array [0..Screen_Height-1,0..1] of Integer;
  End;

Const
  Max_Boundaries = 65528 div SizeOf(TBoundary);

Type
  TBoundaries = Array [0..Max_Boundaries-1] of TBoundary;

Var
  StackError : Boolean;      { Portal stack overflow/underflow flag }
  SStack : Array[0..Max_Boundaries-1] of Integer;       { The stack }
  SSP    : Integer;                             { The stack pointer }
  PBoundaries : ^TBoundaries;

Procedure Clear_Stack;          { Clears the stack }
Procedure SPush (SN : Integer); { Pushes a value SN onto the stack }
Procedure SPop;                 { Popes one value from the stack away }
Function Is_On_Stack (SN : Integer) : Boolean;

Function Allocate_Boundaries : Boolean;
Procedure Deallocate_Boundaries;
Procedure Put_Screen_Boundaries;


Implementation

Procedure Clear_Stack;                  { Clears the stack }
Begin
  SSP := 0
End;

Procedure SPush (SN : Integer);         { Pushes a value SN onto the stack }
Begin
  SStack[SSP] := SN;
  Inc (SSP);
  If SSP = Max_Boundaries then Begin
    StackError := True;                 { Stack overflow }
    Halt
  End
End;

Procedure SPop;                         { Popes one value from the stack away }
Begin
  Dec (SSP);
  If SSP = -1 then Begin
    StackError := True;                 { Stack underflow }
    Halt
  End
End;

Function Is_On_Stack (SN : Integer) : Boolean;
{ Returns TRUE if SN number is in the stack, otherwise returns FALSE }
Var
  I : Word;
  Result : Boolean;
Begin
  Result := False;
  I := SSP;
  While I <> 0 do Begin
    Dec (I);
    If SStack[I] = SN then Begin
      Result := True;
      Break
    End
  End;
  Is_On_Stack := Result
End;

Function Allocate_Boundaries : Boolean;
Begin
  PBoundaries := Nil;
  If MaxAvail >= SizeOf(TBoundaries) then
    GetMem (PBoundaries, SizeOf(TBoundaries));
  Allocate_Boundaries := PBoundaries <> Nil
End;

Procedure Deallocate_Boundaries;
Begin
  If PBoundaries <> nil then
    FreeMem (PBoundaries, SizeOf(TBoundaries))
End;

Procedure Put_Screen_Boundaries;
Var I : Integer;
Begin
  If PBoundaries = Nil then Exit;
  Clear_Stack;
  SPush (-1);
  with PBoundaries^[0] do Begin
    Xmin := 0; Xmax := Screen_Width-1;
    Ymin := 0; Ymax := Screen_Height-1;
    For I := 0 to Screen_Height-1 do Begin
      PBoundaries^[0].Boundary[I,0] := 0;
      PBoundaries^[0].Boundary[I,1] := Screen_Width-1
    End
  End
End;

Begin
  Clear_Stack;
  PBoundaries := Nil;
  StackError := False
End.
