{$G+,N+}                                                { 286+, 287+ }
{$R-,S-}
Unit Vectors;

Interface

                            (* TYPE DEFINITIONS *)

Type
  TVector = Record                      { 3d point/vector definition }
    X, Y, Z : Single;
  End;

                               (* VARIABLES *)

Var
  SinTab,                                       { Sine table }
  CosTab   : Array[0..359] of Double;           { Cosine table }

                         (* VECTOR MATH FUNCTIONS *)

Procedure Vector_Add (Var A, B, C : TVector);
Procedure Vector_Sub (Var A, B, C : TVector);
Function Vector_Length (Var A : TVector) : Double;
Procedure Vector_Unit (Var A : TVector);
Function Dot_Product (Var A, B : TVector) : Double;
Procedure Cross_Product (Var A, B, C : TVector);


Implementation

                         (* VECTOR MATH FUNCTIONS *)

Procedure Vector_Add (Var A, B, C : TVector);
Begin
  C.X := A.X + B.X;
  C.Y := A.Y + B.Y;
  C.Z := A.Z + B.Z
End;

Procedure Vector_Sub (Var A, B, C : TVector);
Begin
  C.X := A.X - B.X;
  C.Y := A.Y - B.Y;
  C.Z := A.Z - B.Z
End;

Function Vector_Length (Var A : TVector) : Double;
Begin
  Vector_Length := Sqrt( Sqr(A.X) + Sqr(A.Y) + Sqr(A.Z) )
End;

Procedure Vector_Unit (Var A : TVector);
Var L : Double;
Begin
  L := Vector_Length (A);
  A.X := A.X / L;
  A.Y := A.Y / L;
  A.Z := A.Z / L
End;

Function Dot_Product (Var A, B : TVector) : Double;
Begin
  Dot_Product := A.X*B.X + A.Y*B.Y + A.Z*B.Z
End;

Procedure Cross_Product (Var A, B, C : TVector);
Begin
  C.X := A.Y*B.Z - A.Z*B.Y;
  C.Y := A.Z*B.X - A.X*B.Z;
  C.Z := A.X*B.Y - A.Y*B.X
End;

Procedure Rotate_Vector_X (Var V : TVector; Angle : Word);
Var
  VV : TVector;
Begin
  VV.Y := (V.Y * CosTab[Angle]  +  V.Z * SinTab[Angle]);
  VV.Z := (V.Z * CosTab[Angle]  -  V.Y * SinTab[Angle]);
  VV.X := V.X;
  V := VV
End;

Procedure Rotate_Vector_Y (Var V : TVector; Angle : Word);
Var
  VV : TVector;
Begin
  VV.Z := (V.Z * CosTab[Angle]  +  V.X * SinTab[Angle]);
  VV.X := (V.X * CosTab[Angle]  -  V.Z * SinTab[Angle]);
  VV.Y := V.Y;
  V := VV
End;

Procedure Rotate_Vector_Z (Var V : TVector; Angle : Word);
Var
  VV : TVector;
Begin
  VV.X := (V.X * CosTab[Angle]  +  V.Y * SinTab[Angle]);
  VV.Y := (V.Y * CosTab[Angle]  -  V.X * SinTab[Angle]);
  VV.Z := V.Z;
  V := VV
End;

Var
  I : Integer;
  Alpha : Double;

Begin
  For I := 0 to 359 do Begin
    Alpha := I*PI/180;
    SinTab[I] := Sin (Alpha);
    CosTab[I] := Cos (Alpha)
  End
End.
